from .util import truncate_id


class Container:
    def __init__(self, cli, container_id):
        self.info = cli.inspect_container(container_id)

    def get(self, *args, **kwargs):
        return self.info.get(*args, **kwargs)

    @property
    def name(self):
        return self.info['Name'].rsplit('/', 1)[1]

    @property
    def descr(self):
        return '%s[%s, %s]' % (self.name, self.info['Config']['Image'], truncate_id(self.info['Id']))

    def __repr__(self):
        return self.descr

    @classmethod
    def all(cls, cli, **kwargs):
        "construct an Image for all images"
        return [
            Container(cli, i['Id'])
            for i in cli.containers(**kwargs)]



from .util import truncate_id


class Image:
    untagged = '<none>:<none>'

    def __init__(self, cli, image_id):
        self.info = cli.inspect_image(image_id)

    def get(self, *args, **kwargs):
        return self.info.get(*args, **kwargs)

    @property
    def tags(self):
        return self.info['RepoTags']

    @property
    def descr(self):
        name = Image.untagged
        if self.tags:
            name = self.tags[0]
        return '%s[%s]' % (name, truncate_id(self.info['Id']))

    def __repr__(self):
        return self.descr

    @classmethod
    def all(cls, cli, **kwargs):
        "construct an Image for all images"
        return [Image(cli, i['Id']) for i in cli.images(**kwargs)]

    @classmethod
    def repotags(cls, cli):
        "returns the set of all repo tags"
        tags = set()
        images = Image.all(cli)
        for image in images:
            for tag in image.info['RepoTags']:
                if tag == Image.untagged:
                    continue
                tags.add(tag)
        return tags



from .container import Container
from .image import Image
from .volume import Volume


def untagged_images_with_usage(cli):
    used_images = {}
    for container in Container.all(cli, all=True):
        used_images[container.get('Image')] = container
    for image in sorted(Image.all(cli, filters={'dangling': True}), key=repr):
        image_id = image.get('Id')
        yield image, used_images.get(image_id)


def dangling_volumes(cli):
    yield from sorted(Volume.all(cli, filters={'dangling': True}), key=repr)


def stopped_containers(cli):
    yield from (container for container in sorted(Container.all(cli, all=True), key=repr)
                if not container.get('State', {}).get('Running'))


def check_latest_image(cli):
    issues = []
    for container in Container.all(cli):
        # determine the tag this container was launched from
        launched_tag = container.get('Config', {}).get('Image')
        if launched_tag is None:
            continue
        current_image = Image(cli, launched_tag)
        running_image_id = container.get('Image')
        current_image_id = current_image.get('Id')
        if running_image_id != current_image_id:
            issues.append('running container %s: launched from outdated version of tag' % (container))
    return issues


def check_untagged_images(cli):
    issues = []
    for image, used_by in untagged_images_with_usage(cli):
        if used_by is None:
            issues.append('image %s: dangling' % (image))
    return issues


def check_dangling_volumes(cli):
    issues = []
    for volume in dangling_volumes(cli):
        issues.append('volumes %s: dangling' % (volume))
    return issues


def check_stopped_containers(cli):
    issues = []
    stopped_count = len(list(stopped_containers(cli)))
    if stopped_count > 0:
        issues.append('%d stopped containers - possibly unneeded' % (stopped_count))
    return issues



from .util import truncate_id


class Volume:
    unlabelled = '<unlabelled>'

    def __init__(self, cli, volume_id):
        self.info = cli.inspect_volume(volume_id)

    def get(self, *args, **kwargs):
        return self.info.get(*args, **kwargs)

    @property
    def name(self):
        return self.info.get('Name')

    @property
    def descr(self):
        name = Volume.unlabelled
        labels = self.info['Labels']
        if type(labels) is dict:
            name = 'named<{}>'.format(labels)
        elif labels:
            name = labels[0]
        return '%s[%s]' % (name, truncate_id(self.name))

    def __repr__(self):
        return self.descr

    @classmethod
    def all(cls, cli, **kwargs):
        "construct an Image for all images"

        def is_named(volume_id):
            info = cli.inspect_volume(volume_id)
            return type(info['Labels']) is dict

        response = cli.volumes(**kwargs)
        volumes = response['Volumes'] or []
        return [Volume(cli, i['Name']) for i in volumes if not is_named(i['Name'])]



import progressbar
import json
import sys
from progressbar import FormatLabel, Percentage, Bar, RotatingMarker


def log_action(s, pfx=''):
    print(pfx + " - %s" % (s))


def log_issue(s, pfx=''):
    print(pfx + " issue: %s" % (s))


def log_warning(s, pfx=''):
    print(pfx + "warning:  %s" % (s))


def log_warnings(header, fix, issues):
    if not issues:
        return
    print(header + ':')
    for issue in issues:
        log_warning(issue, '  ')
    if fix:
        print("  fix: %s" % (fix))


def log_issues(header, fix, issues):
    if not issues:
        return
    print(header + ':')
    for issue in issues:
        log_issue(issue, '  ')
    if fix:
        print("  fix: %s" % (fix))


def log_any_error(fn):
    try:
        fn()
    except Exception as e:
        print("error:  %s" % (str(e)))


def confirm_action(background, question):
    print(background)
    return input('%s [yN] ' % (question)).startswith('y')


def truncate_id(s):
    return s.split(':', 1)[-1][:12]


def print_status_stream(title, stream):
    widgets = [title, FormatLabel(''), ' ', Percentage(), ' ', Bar(), ' ', RotatingMarker()]
    bar = None
    if sys.stderr.isatty():
        bar = progressbar.ProgressBar(widgets=widgets, max_value=255)

    def print_error(status):
        print(status['error'])

    def print_status(status):
        progress = status.get('progressDetail')
        if progress:
            widgets[1] = FormatLabel("%12s" % (status['status']))
            prog = int(round(255 * ((progress['current'] / progress['total']))))
            if bar is not None:
                bar.update(prog)

    def print_unknown(status):
        print(status)

    for line in stream:
        try:
            status = json.loads(line.decode('utf8'))
        except json.JSONDecodeError:
            continue
        if 'error' in status:
            print_error(status)
        elif 'status' in status:
            print_status(status)
        else:
            print_unknown(status)


def make_registration_decorator():
    """
    returns a (decorator, list). any function decorated with
    the returned decorator will be appended to the list
    """
    registered = []

    def _register(fn):
        registered.append(fn)
        return fn

    return _register, registered



import argparse
import sys
import re

from docker import APIClient as Client
from docker.errors import APIError
from fnmatch import fnmatch

from .image import Image
from .container import Container
from .util import print_status_stream, make_registration_decorator, \
    log_action, log_any_error, log_issue, log_issues, log_warnings, \
    confirm_action
from .check import check_latest_image, check_dangling_volumes, \
    check_untagged_images, untagged_images_with_usage, \
    check_stopped_containers, stopped_containers, \
    dangling_volumes

register_command, command_fns = make_registration_decorator()


@register_command
def pull_all(args):
    "pull all images"
    def status_title(tag, pad=None):
        title = 'pull %s' % (tag)
        if pad:
            title = '%*s' % (pad, title)
        return title

    def pull_tags(tags):
        pad = max(len(status_title(t)) for t in tags)
        for tag in sorted(tags):
            log_action("pulling tag: %s" % (tag))
            try:
                print_status_stream(
                    status_title(tag, pad),
                    cli.pull(tag, stream=True))
            except APIError as err:
                print(err)

    cli = Client()
    tags = Image.repotags(cli)
    if tags:
        pull_tags(tags)


@register_command
def kill_all(args):
    "kill all running containers"
    cli = Client()
    to_kill = list(sorted(Container.all(cli), key=repr))
    if not to_kill:
        return
    background = ['The following running containers will be killed:\n']
    background += [' - %s\n' % (container) for container in to_kill]
    if not args.force and not confirm_action(
            ''.join(background), 'Kill containers?'):
        return
    for container in to_kill:
        log_action("killing container: %s" % (container))
        log_any_error(lambda: cli.kill(container.get('Id')))


@register_command
def rm_stopped(args):
    "remove all containers which are not running"
    cli = Client()
    to_remove = list(stopped_containers(cli))
    if not to_remove:
        return
    background = ['The following stopped containers will be removed:\n']
    background += [' - %s\n' % (container) for container in to_remove]
    if not args.force and not confirm_action(
            ''.join(background), 'Remove containers?'):
        return
    for container in to_remove:
        log_action("removing container: %s" % (container))
        log_any_error(lambda: cli.remove_container(container.get('Id')))


@register_command
def rmv_dangling(args):
    "remove all dangling volumes"
    cli = Client()
    to_remove = list(dangling_volumes(cli))
    if not to_remove:
        return
    background = ['The following dangling volumes will be removed:\n']
    background += [' - %s\n' % (volume) for volume in to_remove]
    if not args.force and not confirm_action(
            ''.join(background), 'Remove volumes?'):
        return
    for volume in to_remove:
        log_action("removing dangling volume: %s" % (volume))
        cli.remove_volume(volume.name)


@register_command
def rmi_dangling(args):
    "remove all dangling (untagged) images"
    cli = Client()
    to_remove = []
    for image, used_by in untagged_images_with_usage(cli):
        if used_by:
            log_issue("not removing image: %s (in use by %s)" % (image, used_by))
        else:
            to_remove.append(image)
    if not to_remove:
        return
    background = ['The following dangling images will be removed:\n']
    background += [' - %s\n' % (image) for image in to_remove]
    if not args.force and not confirm_action(
            ''.join(background), 'Remove images?'):
        return
    for image in to_remove:
        log_action("removing dangling image: %s" % (image))
        log_any_error(lambda: cli.remove_image(image.get('Id')))


@register_command
def scrub(args):
    "remove all stopped containers, dangling images and volumes"
    rm_stopped(args)
    rmi_dangling(args)
    rmv_dangling(args)


def setup_force(subparser):
    subparser.add_argument('--force', help='don\'t ask to confirm', action='store_true')

kill_all.setup = setup_force
rm_stopped.setup = setup_force
rmi_dangling.setup = setup_force
rmv_dangling.setup = setup_force
scrub.setup = setup_force


def match_iterator_glob_or_regexp(args, iterator, apply_fn):
    """
    returns the matching objects from `iterator`, using fnmatch
    unless args.e is set, in which case regular expression is used.
    `apply_fn` is applied to each object, returning a string to
    check match against.
    """
    if args.e:
        r = re.compile(args.pattern)

        def matcher(s):
            return r.match(s)
    else:
        def matcher(s):
            return fnmatch(s, args.pattern)

    for obj in iterator:
        match = apply_fn(obj)
        if matcher(match):
            yield obj


@register_command
def rm_matching(args):
    "remove containers whose name matches `pattern'"
    cli = Client()
    to_remove = list(
        match_iterator_glob_or_regexp(
            args,
            stopped_containers(cli),
            lambda c: c.name))
    if not to_remove:
        return
    background = ['The following containers will be deleted:\n']
    for container in sorted(to_remove, key=repr):
        background.append(' - %s\n' % (container))
    if not args.force and not confirm_action(
            ''.join(background), 'Delete matching containers?'):
        return
    for container in to_remove:
        log_action("removing container via tag: %s" % (container))
        log_any_error(lambda: cli.remove_container(container.get('Id')))


@register_command
def rmi_matching(args):
    "remove images which have tags matching `pattern'"
    cli = Client()

    def all_image_tags():
        for image in Image.all(cli):
            for tag in image.tags:
                yield tag
    to_remove = list(
        match_iterator_glob_or_regexp(
            args,
            all_image_tags(),
            lambda t: t))
    if not to_remove:
        return
    background = ['Images with the following tags will be deleted:\n']
    for tag in sorted(to_remove):
        background.append(' - %s\n' % (tag))
    if not args.force and not confirm_action(
            ''.join(background), 'Delete matching images?'):
        return
    for tag in to_remove:
        log_action("removing image via tag: %s" % (tag))
        log_any_error(lambda: cli.remove_image(tag))


@register_command
def doctor(args):
    "check for common issues"
    cli = Client()
    log_issues("containers running from old version of tag", "restart containers", check_latest_image(cli))
    log_issues("dangling volumes", "wrfy rmv-dangling", check_dangling_volumes(cli))
    log_issues("dangling images", "wrfy rmi-dangling", check_untagged_images(cli))
    log_warnings("stopped containers", "wrfy rm-stopped", check_stopped_containers(cli))


def setup_globrm(subparser):
    subparser.add_argument('pattern', help='pattern (by default, glob)')
    subparser.add_argument('-e', help='treat pattern as regular expression', action='store_true')
    subparser.add_argument('--force', help='don\'t ask to confirm', action='store_true')
rm_matching.setup = setup_globrm
rmi_matching.setup = setup_globrm


def version():
    import pkg_resources
    version = pkg_resources.require("wrfy")[0].version
    print('''\
wrfy, version %s

Copyright © 2016 Grahame Bowland
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.''' % (version))
    sys.exit(0)


def usage(parser):
    parser.print_usage()
    sys.exit(0)


def commands():
    for fn in command_fns:
        name = fn.__name__.replace('_', '-')
        yield name, fn, getattr(fn, 'setup', None), fn.__doc__


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--version', action='store_true',
        help='print version and exit')

    subparsers = parser.add_subparsers(dest='name')
    for name, fn, setup_fn, help_text in sorted(commands()):
        subparser = subparsers.add_parser(name, help=help_text)
        subparser.set_defaults(func=fn)
        if setup_fn is not None:
            setup_fn(subparser)
    args = parser.parse_args()
    if args.version:
        version()
    if 'func' not in args:
        usage(parser)
    args.func(args)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wrfy']

package_data = \
{'': ['*']}

install_requires = \
['docker>=2.7.0,<3.0.0', 'progressbar2>=3.55.0,<4.0.0']

entry_points = \
{'console_scripts': ['wrfy = wrfy.cli:main']}

setup_kwargs = {
    'name': 'wrfy',
    'version': '0.9.2',
    'description': 'Simple bulk operations for Docker',
    'long_description': "# `wrfy` /wharfie/\n\n> *minimal CLI tool to smooth your docker local dev experience*\n\n`wrfy` provides a dozen or so commands to automate common operations on a docker development or CI host. Want to pull all images? Delete images matching a regexp? Clean up dangling volumes? You've come to the right place.\n\n# installation\n\n```$ pip3 install wrfy```\n\n# `wrfy` commands\n\n## doctor\n\n`wrfy doctor` will check your docker host for common issues. The checks are:\n\n  - **containers running from an old version of the image they were launched from**. for example, \n    if you were to do `docker run -it alpine:latest /bin/sh`, leave that\n\tcontainer going, and then pull a newer version of `alpine:latest`, `wrfy`\n\twill let you know that your Alpine container is running from an old\n\timage.\n  - **dangling volumes**. dangling volumes, which are not attached to a container.\n  - **dangling images**. dangling images, which do not have a tag.\n  - **stopped containers**. docker hosts can build up a large number of stopped containers\n    whose purpose was ephemeral.\n\nEach check suggests a `wrfy` tool to address each particular issue identified.\n\n## kill-all\n\n`wrfy kill-all` will kill all running containers.\nIt asks for confirmation, unless `--force` is passed as an argument.\n\n## pull-all\n\n`wrfy pull-all` will pull all images present on the docker host. This is very useful when you want to make sure everything is up to date.\n\n## rm-matching\n\n`wrfy rm-matching <pattern>` will remove containers matching the provided glob pattern.\nIf `-e` is passed, the pattern is interpreted as a regular expression.\nIt asks for confirmation, unless `--force` is passed as an argument.\n\nA useful command is `wrfy rm-matching -e '^[a-z]+_[a-z]+$'`, which will remove all containers with a name\ncomprised of two words seperated by an underscore. This will match containers with names automatically\ngenerated by docker.\n\n## rm-stopped\n\n`wrfy rm-stopped` will remove all containers which are not running. It is somewhat of a blunt instrument,\nyou might want to use `rm-matching` instead.\nIt asks for confirmation, unless `--force` is passed as an argument.\n\n## rmi-dangling\n\n`wrfy rmi-dangling` will remove all dangling images - images which haven't got a name. \nIt asks for confirmation, unless `--force` is passed as an argument.\n\n## rmi-matching\n\n`wrfy rmi-matching <pattern>` will remove images matching the provided glob pattern.\nIf `-e` is passed, the pattern is interpreted as a regular expression.\nIt asks for confirmation, unless `--force` is passed as an argument.\n\nA useful command to clean up after `docker-compose` is \n`wrfy rmi-matching -e '^[a-z]+_[a-z]+:latest$'`,\nwhich will remove all containers with a name comprised of two words\nseperated by an underscore, and a tag of `latest`. Such images\nprobably came from `docker-compose`.\n\n## rmv-dangling\n\n`wrfy rmv-dangling` will remove all dangling volumes - volumes not attached to any container.\nIt asks for confirmation, unless `--force` is passed as an argument.\n\n## scrub\n\n`wrfy scrub` chains together `rm_stopped`, `rmi_dangling`, and `rmv_dangling`.\nIt asks for confirmation at each stage, unless `--force` is passed as an argument.\n\n",
    'author': 'Grahame Bowland',
    'author_email': 'grahame@oreamnos.com.au',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
